# Copyright (c) 2007 Thomas Lotze
# See also LICENSE.txt

import unittest
import doctest
from zope.testing.doctest import DocFileSuite

import pkg_resources
import setuptools.tests.test_resources


def make_dist(filename, depends=""):
    metadata = setuptools.tests.test_resources.Metadata(("depends.txt",
                                                         depends))
    return pkg_resources.Distribution.from_filename(filename,
                                                    metadata=metadata)


def make_working_set(*dists):
    ws = pkg_resources.WorkingSet([])
    for dist in dists:
        ws.add(dist)
    return ws


def _sprint(value):
    if isinstance(value, set):
        return "set([%s])" % ", ".join(repr(elem) for elem in sorted(value))
    if isinstance(value, dict):
        return "{%s}" % ", ".join("%r: %s" % (key, _sprint(value))
                                  for key, value in sorted(value.iteritems()))
    return repr(value)


def sprint(value):
    print _sprint(value)


def sort_specs(specs):
    return sorted(specs, cmp=lambda a, b: cmp(a.project_name, b.project_name))


def setUp(test):
    test.globs.update(dict(
        make_dist=make_dist,
        make_working_set=make_working_set,
        sprint=sprint,
        sort_specs=sort_specs,
        ))


def test_suite():
    return unittest.TestSuite([
        DocFileSuite(filename,
                     package="tl.eggdeps",
                     setUp=setUp,
                     optionflags=doctest.NORMALIZE_WHITESPACE |
                                 doctest.ELLIPSIS,
                     )
        for filename in (
        "graph.txt",
        )])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# Copyright (c) 2007 Thomas Lotze
# See also LICENSE.txt

import sys
import pkg_resources


class Graph(dict):
    """A graph of egg dependencies.
    """

    def __init__(self,
                 working_set=None,
                 show=lambda name: True,
                 follow=lambda name: True,
                 extras=True,
                 ):
        self.working_set = working_set or pkg_resources.WorkingSet()
        self.show = show
        self.follow = follow
        self.extras = extras
        self.roots = ()
        self.show_dist = lambda spec: show(spec.project_name)

    def from_specifications(self, *specifications):
        requirements = set(pkg_resources.parse_requirements(specifications))
        self.roots = self.names(requirements)

        for req in filter(self.show_dist, requirements):
            self.add_requirement(req)

    def add_requirement(self, req):
        node = self.setdefault(req.project_name, Node(self, req))
        if not (node.follow and node.dist):
            return

        if not node:
            for dep in self.names(node.dist.requires()):
                node[dep] = set()

        new_reqs = set(node.dist.requires())
        if self.extras:
            plain_names = self.names(new_reqs)
            for extra in req.extras:
                extra_reqs = node.dist.requires((extra,))
                new_reqs.update(extra_reqs)
                for dep in self.names(extra_reqs) - plain_names:
                    node.setdefault(dep, set()).add(extra)

        new_reqs -= node.requires
        node.requires |= new_reqs
        for req in filter(self.show_dist, new_reqs):
            self.add_requirement(req)

    def from_working_set(self):
        ws = filter(self.show_dist, self.working_set)
        ws_names = self.names(ws)
        self.roots = ws_names.copy()

        for dist in ws:
            node = self[dist.project_name] = Node(self, dist)
            if not node.follow:
                continue

            for dep in self.names(dist.requires()) & ws_names:
                node[dep] = set()

            plain_names = self.names(dist.requires())
            if self.extras:
                for extra in dist.extras:
                    for dep in (self.names(dist.requires((extra,)))
                                & ws_names - plain_names):
                        node.setdefault(dep, set()).add(extra)

            self.roots -= set(node)

    def names(self, collection):
        return set(filter(self.show, (x.project_name for x in collection)))


class Node(dict):
    """A graph node representing and egg and its dependencies.
    """

    def __init__(self, graph, spec):
        self.name = spec.project_name
        self.graph = graph
        self.requires = set()
        self.follow = self.graph.follow(self.name)

        # Find a distribution matching the spec in the working set. Search
        # even if the spec is already a distribution to make sure it's active.
        if isinstance(spec, pkg_resources.Distribution):
            spec = spec.as_requirement()
        try:
            self.dist = self.graph.working_set.find(spec)
        except pkg_resources.VersionConflict:
            self.dist = None



# Copyright (c) 2007 Thomas Lotze
# See also LICENSE.txt

import re
import optparse

import tl.eggdeps.graph
import tl.eggdeps.dot
import tl.eggdeps.plaintext


def eggdeps():
    parser = optparse.OptionParser("usage: %prog [options] [specifications]")
    parser.add_option("-i", "--ignore",
                      dest="ignore", action="append", default=[],
                      help="project names to ignore")
    parser.add_option("-I", "--re-ignore",
                      dest="re_ignore", action="append", default=[],
                      help="regular expression for project names to ignore")
    parser.add_option("-e", "--dead-end",
                      dest="dead_ends", action="append", default=[],
                      help="names of projects whose dependencies to ignore")
    parser.add_option("-E", "--re-dead-end",
                      dest="re_dead_ends", action="append", default=[],
                      help="regular expression for project names "
                           "whose dependencies to ignore")
    parser.add_option("-n", "--no-extras",
                      dest="extras", action="store_false", default=True,
                      help="always omit extra dependencies")
    parser.add_option("-d", "--dot",
                      dest="dot", action="store_true", default=False,
                      help="produce a dot graph")
    parser.add_option("-c", "--cluster",
                      dest="cluster", action="store_true", default=False,
                      help="in a dot graph, cluster direct dependencies "
                           "of each root distribution")
    options, specifications = parser.parse_args()

    show = unmatcher(options.ignore, options.re_ignore)
    follow = unmatcher(options.dead_ends, options.re_dead_ends)

    graph = tl.eggdeps.graph.Graph(show=show,
                                   follow=follow,
                                   extras=options.extras,
                                   )

    if specifications:
        graph.from_specifications(*specifications)
    else:
        graph.from_working_set()

    if options.dot:
        tl.eggdeps.dot.print_dot(graph, cluster=options.cluster)
    else:
        tl.eggdeps.plaintext.print_graph(graph)


def unmatcher(names, patterns):
    names = set(names)
    compiled_patterns = [re.compile(pattern) for pattern in patterns]

    def unmatch(name):
        if name in names:
            return False

        for re in compiled_patterns:
            if re.search(name):
                return False

        return True

    return unmatch



# Copyright (c) 2007 Thomas Lotze
# See also LICENSE.txt


def print_dot(graph, cluster=False):
    direct_deps = set()
    for name in graph.roots:
        direct_deps.update(graph[name])

    print "digraph {"

    for node in graph.itervalues():
        node_options = dict(label=node.name)

        def fill(color):
            node_options["style"] = "filled"
            node_options["fillcolor"] = color
        if node.name in direct_deps:
            fill("yellow")
        if node.name in graph.roots:
            fill("green")
        if not node.follow:
            fill("lightgrey")
        if not node.dist:
            fill("red")

        print node.name.replace(".", "_") + format_options(node_options)

    if cluster:
        for i, cluster in enumerate(yield_clusters(graph)):
            print "subgraph cluster_%s {" % i
            for name in cluster:
                print name.replace(".", "_")
            print "}"

    for node in graph.itervalues():
        for dep, extras in node.iteritems():
            edge_options = {}
            if extras:
                edge_options["color"] = "lightgrey"

            print "%s -> %s%s" % (node.name.replace(".", "_"),
                                  dep.replace(".", "_"),
                                  format_options(edge_options))

    print "}"


def format_options(options):
    if not options:
        return ""

    return " [%s]" % ", ".join('%s="%s"' % item
                               for item in options.iteritems())


def yield_clusters(graph):
    clusters = [set(graph[name]).union((name,)) for name in graph.roots]
    while clusters:
        cluster = clusters.pop()
        for other in reversed(clusters):
            if cluster & other:
                cluster |= other
                clusters.remove(other)
        yield cluster



#!/usr/bin/env python
# -*- coding: latin-1 -*-
#
# Copyright (c) 2007 Thomas Lotze
# See also LICENSE.txt

"""Eggs grow on trees. Dependency trees.
"""

import os, os.path
import glob
from setuptools import setup, find_packages


project_path = lambda *names: os.path.join(os.path.dirname(__file__), *names)


def include_tree(dest, source):
    source_len = len(source)
    for dirpath, dirnames, filenames in os.walk(source):
        yield (dest + dirpath[source_len:],
               [os.path.join(dirpath, fn) for fn in filenames])
        if ".svn" in dirnames:
            dirnames.remove(".svn")


longdesc = open(project_path("README.txt")).read()

data_files = [("", glob.glob(project_path("*.txt")))]

entry_points = {
    "console_scripts": [
    "eggdeps = tl.eggdeps.cli:eggdeps",
    ],
    }

install_requires = [
    "setuptools",
    ]

extras_require = {
    "test": ["zope.testing"],
    }

tests_require = [
    "zope.testing",
    ]

classifiers = [
    "Development Status :: 3 - Alpha",
    ]

setup(name="tl.eggdeps",
      version="0.1",
      description=__doc__.strip(),
      long_description=longdesc,
      keywords="egg eggs dependencies dependency graph tree",
      classifiers=classifiers,
      author="Thomas Lotze",
      author_email="thomas@thomas-lotze.de",
      url="http://www.thomas-lotze.de/en/software/eggdeps/",
      license="ZPL 2.1",
      packages=find_packages(),
      entry_points=entry_points,
      install_requires=install_requires,
      extras_require=extras_require,
      tests_require=tests_require,
      include_package_data=True,
      data_files=data_files,
      test_suite="tl.eggdeps.tests.test_suite",
      )



# Copyright (c) 2007 Thomas Lotze
# See also LICENSE.txt


def print_subgraph(name, graph, mount_points, extras=None, path=()):
    print_tree = path == mount_points[name]
    root = graph[name]

    line = len(path) * "    "
    if root.dist:
        line += name
    else:
        line += "(%s)" % name
    if extras:
        line += " [%s]" % ','.join(sorted(extras))
    if not print_tree and root:
        line += " ..."
    if not root.follow:
        line += " *"
    print line

    if not print_tree:
        return

    for dep, extras in sorted(
        root.iteritems(),
        cmp=lambda (a, b), (c, d): cmp(sorted(b), sorted(d)) or cmp(a, c)):
        print_subgraph(dep, graph, mount_points, extras, path + (name,))


def find_mount_point(name, graph, mount_points, plain_mounts,
                     path=(), is_plain=True):
    # prefer paths without extra dependencies
    if name in plain_mounts and not is_plain:
        return

    mount_point = mount_points.get(name)
    if mount_point is not None:
        # prefer short paths
        if len(path) > len(mount_point):
            return
        # prefer alphabetically first among paths of equal length
        if len(path) == len(mount_point) and path > mount_point:
            return

    mount_points[name] = path
    if is_plain:
        plain_mounts.add(name)

    path += (name,)
    for dep, extras in graph[name].iteritems():
        find_mount_point(dep, graph, mount_points, plain_mounts,
                         path, is_plain and not extras)


def print_graph(graph):
    mount_points = {}
    plain_mounts = set()
    for name in graph.roots:
        find_mount_point(name, graph, mount_points, plain_mounts)

    for name in sorted(graph.roots):
        print_subgraph(name, graph, mount_points)
